/* Copyright (C) 2003-2015 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Martin Schwidefsky <schwidefsky@de.ibm.com>, 2003.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include <endian.h>
#include <errno.h>
#include <sysdep.h>
#include <lowlevellock.h>
#include <pthread.h>
#include <pthreadP.h>
#include <atomic.h>

#include <shlib-compat.h>
#include <kernel-features.h>
#include <stap-probe.h>


/* See __pthread_cond_wait for a high-level description of the algorithm.  */
int
__pthread_cond_signal (pthread_cond_t *cond)
{
  unsigned int wseq, ssent;

  /* MUTEX might get modified concurrently, but relaxed memory order is fine:
     In case of a shared condvar, the field will be set to value ~0l during
     initialization of the condvar (which happens before any signaling) and
     is immutable afterwards; otherwise, the field will never be set to a
     value of ~0l.  */
  int pshared = (atomic_load_relaxed (&cond->__data.__mutex) == (void *) ~0l)
		? LLL_SHARED : LLL_PRIVATE;

  LIBC_PROBE (cond_signal, 1, cond);

  /* Load the waiter sequence number, which represents our relative ordering
     to any waiters.  Also load the number of signals sent so far.
     We do not need stronger MOs for both loads nor an atomic snapshot of both
     because:
     1) We can pick any position that is allowed by external happens-before
        constraints.  In particular, if another __pthread_cond_wait call
        happened before us, this waiter must be eligible for being woken by
        us.  The only way do establish such a happens-before is by signaling
        while holding the mutex associated with the condvar and ensuring that
        the signal's critical section happens after the waiter.  Thus, the
        mutex ensures that we see this waiter's wseq increase.
     2) Once we pick a position, we do not need to communicate this to the
        program via a happens-before that we set up: First, any wake-up could
        be a spurious wake-up, so the program must not interpret a wake-up as
        an indication that the waiter happened before a particular signal;
        second, a program cannot detect whether a waiter has not yet been
        woken (i.e., it cannot distinguish between a non-woken waiter and one
        that has been woken but hasn't resumed execution yet), and thus it
        cannot try to deduce that a signal happened before a particular
        waiter.
     3) The load of WSEQ does not need to constrain which value we load for
        SIGNALS_SENT: If we read an older value for SIGNALS_SENT (compared to
        what would have been current if we had an atomic snapshot of both),
        we might send a signal even if we don't need to; thus, we just get a
        spurious wakeup.  If we read a more recent value (again, compared to
        an atomic snapshot), then some other signal interfered and might have
        taken "our" position in the waiter/wake-up sequence; thus the waiters
        we had to wake will get woken either way.
     Note that we do not need to check whether the generation changed
     concurrently: If it would change, we would just skip any signaling
     because we could have been positioned in the earlier generation -- all
     the waiters we would have to wake will have been woken during the
     quiescence period.  Thus, at worst we will cause one additional spurious
     wake-up if we don't detect this.  */
  wseq = atomic_load_relaxed (&cond->__data.__wseq);
  ssent = atomic_load_relaxed (&cond->__data.__signals_sent);
  do
    {
      /* If we don't have more waiters than signals or are in a quiescence
         period, just return because all waiters we must wake have been or
         will be woken.  See above for further details.  */
      if (ssent >= wseq || wseq >= __PTHREAD_COND_WSEQ_THRESHOLD)
	return 0;
    }
  /* Using a CAS loop here instead of a fetch-and-increment avoids one source
     of spurious wake-ups, namely several signalers racing to wake up a fewer
     number of waiters and thus also waking subsequent waiters spuriously.
     The cost of this is somewhat more contention on SIGNALS_SENT on archs
     that offer atomic fetch-and-increment.
     TODO Relaxed MO is sufficient here.
   */
  while (!atomic_compare_exchange_weak_relaxed (&cond->__data.__signals_sent,
						&ssent, ssent + 1));

  /* XXX Could we skip the futex_wake if not necessary (eg, if there are just
     spinning waiters)?  This would need additional communication but could it
     be more efficient than the kernel-side communication?  Should we spin for
     a while to see if our signal was consumed in the meantime?  */
  lll_futex_wake (&cond->__data.__signals_sent, 1, pshared);

  return 0;
}

versioned_symbol (libpthread, __pthread_cond_signal, pthread_cond_signal,
		  GLIBC_2_3_2);
